package work.xuye.sink.handler;

import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.springframework.messaging.Message;
import org.springframework.messaging.MessageHeaders;
import org.springframework.stereotype.Component;
import org.springframework.util.Assert;
import org.springframework.util.DigestUtils;
import org.springframework.util.ObjectUtils;
import work.xuye.common.constant.MessageConstants;
import work.xuye.common.constant.StageConstants;
import work.xuye.common.db.entity.vo.SinkParams;
import work.xuye.common.db.entity.vo.TableTemplate;
import work.xuye.common.db.service.TaskManager;
import work.xuye.common.dto.TaskVO;
import work.xuye.common.enums.ProcessMode;
import work.xuye.common.enums.ProcessReason;
import work.xuye.common.enums.ResourceStatus;
import work.xuye.common.enums.Status;
import work.xuye.common.service.UrlMD5Service;
import work.xuye.common.spel.CustomFunction;
import work.xuye.common.spel.SpringExpressionLanguageEvaluator;
import work.xuye.sink.service.JdbcService;

import java.util.HashMap;
import java.util.Map;

/**
 * @author xuye
 * @since 2023/2/19 17:20
 **/
@Slf4j
@Component
@RequiredArgsConstructor
public class ItemHandler {
    private final JdbcService jdbcService;
    private final UrlMD5Service urlMD5Service;
    private final SpringExpressionLanguageEvaluator evaluator;
    private final CustomFunction customFunction;

    private final TaskManager taskManager;

    private boolean needHandle(Message<String> message) {
        MessageHeaders headers = message.getHeaders();
        String currentStage = (String) headers.get(MessageConstants.CURRENT_STAGE);
        if (StageConstants.END.equals(currentStage)) {
            log.debug("@@@ 此消息已经处于结束状态，无需处理");
            return false;
        }
        return true;
    }


    public void handle(Message<String> message) {
        try {
            this.doHandle(message);
        } catch (Exception e) {
            // 如果这一步出错，需要清除seedUrl缓存，避免上一步的缓存仍然存在，影响执行到本类的下一次重试
            this.cleanCache(message);
            throw new RuntimeException(e);
        }
    }

    private void cleanCache(Message<String> message) {
        String seedUrl = (String) message.getHeaders().get(MessageConstants.SEED_URL);
        urlMD5Service.removeUrlMD5(seedUrl);
    }


    public void doHandle(Message<String> message) {
        if (!needHandle(message)) {
            return;
        }
        TaskVO taskVO = taskManager.getTaskInfoByTaskId((Integer) message.getHeaders().get(MessageConstants.TASK_ID));
        //相关字段放这里
        String md5Digest = DigestUtils.md5DigestAsHex(message.getPayload().getBytes());
        TableTemplate template = taskVO.getTemplate().getTableTemplate();
        // 根据消息体，和映射关系，使用SpEL表达式，得到计算结果
        HashMap<String, Object> expressionResultMap = evaluator.evaluate(taskVO.getMapping().getTableMapping(), message);
        // 如果不能通过校验，则抛出异常
        this.validate(expressionResultMap, template);
        // 将计算结果合并到模板中，此刻的模板是拥有值的
        this.mergeValueToTemplate(taskVO, template, expressionResultMap);
        // 根据MD5判断是否需要处理
        ProcessReason reason = this.getReason(taskVO, template, md5Digest);
        if (reason == ProcessReason.unchanged) {
            return;
        }
        this.log(message, template, reason);
        String dataSourceName = taskVO.getTask().getSinkParams().getDataSourceName();
        //判断数据是否存在
        boolean itemExist = this.itemExist(taskVO.getTask().getSinkParams(), template);
        //如果已经存在了，就判断是否需要更新，如果需要更新就更新
        if (itemExist) {
            this.tryUpdate(taskVO.getTask().getSinkParams(), expressionResultMap, md5Digest, template);
            //如果不存在，就插入
        } else {
            this.tryInsert(dataSourceName, template);
        }
        // 必须在方法出栈前，将md5放入缓存，因为Spring Cloud Stream默认会重试三次，如果提前放入缓存了，会导致首次重试被判定为无需处理
        urlMD5Service.put(template.getUniqueField().getValue().toString(), md5Digest);
        return;
    }


    private void validate(HashMap<String, Object> expressionResultMap, TableTemplate tableTemplate) {
        if (!expressionResultMap.containsKey(tableTemplate.getUniqueField().getFieldName())) {
            throw new RuntimeException("uniqueField [" + tableTemplate.getUniqueField().getFieldName() + "] not found");
        }
    }

    private void mergeValueToTemplate(TaskVO taskVO, TableTemplate template, HashMap<String, Object> expressionResultMap) {
        // 唯一字段一定要使用计算结果
        template.getUniqueField().setValue(expressionResultMap.get(template.getUniqueField().getFieldName()));

        // 将计算结果，填充/覆盖模板中的字段
        for (TableTemplate.Field field : template.getOtherFields()) {
            String fieldName = field.getFieldName();
            if (expressionResultMap.containsKey(fieldName)) {
                Object value = expressionResultMap.get(fieldName);
                field.setValue(value);
            }
        }

        // 以任务的表名为准，覆盖模板的表名
        if (!ObjectUtils.isEmpty(taskVO.getTask().getSinkParams().getTableName())) {
            template.setTableName(taskVO.getTask().getSinkParams().getTableName());
        }
    }

    private void log(Message<String> message, TableTemplate tableTemplate, ProcessReason reason) {
        log.info("@@ [{} {}] 开始处理: " + MessageConstants.TASK_TRACE_ID + ": [{}], "
                        + MessageConstants.SOURCE_TRACE_ID + ": [{}], "
                        + MessageConstants.TRANSFORMER_TRACE_ID + ": [{}], "
                        + "{}: [{}]",
                reason.getEmoji(),
                reason.name(),
                tableTemplate.getUniqueField().getFieldName(),
                tableTemplate.getUniqueField().getValue(),
                message.getHeaders().get(MessageConstants.TASK_TRACE_ID),
                message.getHeaders().get(MessageConstants.SOURCE_TRACE_ID),
                message.getHeaders().get(MessageConstants.TRANSFORMER_TRACE_ID));
    }

    private ProcessReason getReason(TaskVO taskVO, TableTemplate template, String md5) {

        String url = template.getUniqueField().getValue().toString();
        ResourceStatus resourceStatus = urlMD5Service.getResourceStatus(url, md5);

        ProcessMode processMode = taskVO.getTask().getProcessMode();
        if (ResourceStatus.UNCHANGED == resourceStatus && ProcessMode.NORMAL == processMode) {
            log.debug("@@ 数据未更新，无需继续处理");
            return ProcessReason.unchanged;
        } else {
            if (ProcessMode.DEBUG == processMode && ResourceStatus.UNCHANGED == resourceStatus) {
                log.debug("@@ 数据未更新，但当前为调试模式，需继续进行处理");
                return ProcessReason.debug_mode;
            } else if (ResourceStatus.UNKNOWN == resourceStatus) {
                log.debug("@@ 数据状态未知，需进行处理");
                return ProcessReason.cache_not_exist;
            } else if (ResourceStatus.UPDATED == resourceStatus) {
                log.debug("@@ 数据相比缓存中上次记录已更新，需进行处理");
                return ProcessReason.resource_changed;
            }
        }
        return ProcessReason.other;
    }

    private boolean itemExist(SinkParams sinkParams, TableTemplate tableTemplate) {
        String selectExist = SqlGenerator.generateExistedSql(tableTemplate);
        return jdbcService.isExist(sinkParams.getDataSourceName(), selectExist);
    }


    private void tryUpdate(SinkParams sinkParams, HashMap<String, Object> resultMap, String md5Digest, TableTemplate tableTemplate) {
        boolean updateConfigEnabled = sinkParams.getCheckUpdate().getStatus().equals(Status.on);
        if (updateConfigEnabled) {
            String dataSourceName = sinkParams.getDataSourceName();
            Assert.notNull(dataSourceName, "dataSourceName is null");

            Map<String, Object> item = jdbcService.getOne(dataSourceName, SqlGenerator.generateFetchSql(tableTemplate));
            SinkParams.CheckUpdate checkUpdateConfig = sinkParams.getCheckUpdate();
            String fieldName = checkUpdateConfig.getFieldName();
            Object dbValue = item.get(fieldName);
            if (ObjectUtils.isEmpty(dbValue)) {
                throw new RuntimeException("根据 [" + fieldName + "] 来判断数据是否需要更新，但是数据库不存在该字段");
            }

            Object nowValue = resultMap.get(fieldName);
            if (!ObjectUtils.isEmpty(checkUpdateConfig.getJsonPath())) {
                dbValue = customFunction.jsonExtract(dbValue.toString(), checkUpdateConfig.getJsonPath());
                nowValue = customFunction.jsonExtract(nowValue.toString(), checkUpdateConfig.getJsonPath());
            } else {
                throw new RuntimeException("根据 [" + fieldName + "] 来判断数据是否需要更新，但是没有配置jsonPath");
            }
            if (dbValue.equals(nowValue)) {
                log.info("@ ✅数据已存在于MySQL，且不需要更新");
                urlMD5Service.put(tableTemplate.getUniqueField().getValue().toString(), md5Digest);
                return;
            } else {
                String update = SqlGenerator.generateUpdateSql(tableTemplate);
                jdbcService.update(dataSourceName, update);
                log.info("@ \uD83D\uDD04✅数据已存在于MySQL，已经更新数据");
            }
        } else {
            log.info("@ ⏭️该任务未开启更新功能，不会进行更新处理");
        }
    }


    private void tryInsert(String dataSourceName, TableTemplate template) {
        jdbcService.update(dataSourceName, SqlGenerator.generateInsertSql(template));
        log.info("@ \uD83D\uDCBE 数据不存在于MySQL，已插入数据");
    }


}