package work.xuye.source.handler;

import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.springframework.messaging.Message;
import org.springframework.messaging.support.MessageBuilder;
import org.springframework.stereotype.Component;
import org.springframework.util.DigestUtils;
import work.xuye.common.constant.MessageConstants;
import work.xuye.common.constant.StageConstants;
import work.xuye.common.db.service.TaskManager;
import work.xuye.common.dto.HttpRequestParams;
import work.xuye.common.dto.HttpRes;
import work.xuye.common.enums.ProcessMode;
import work.xuye.common.enums.ResourceStatus;
import work.xuye.common.service.UrlMD5Service;
import work.xuye.common.utils.ID;
import work.xuye.source.request.RequestClient;

/**
 * @author xuye
 * @since 2023/2/19 16:37
 **/
@Slf4j
@Component
@RequiredArgsConstructor
public class SourceHandler {

    private final TaskManager taskManager;
    private final RequestClient requestClient;
    private final UrlMD5Service urlMD5Service;


    public Message<HttpRes> handle(Message<HttpRequestParams> message) {


        // 先根据req拿到res，包含从缓存中检查该URL的状态
        HttpRes res = null;
        try {
            res = this.doHandle(message);
        } catch (Exception e) {
            log.error("@@ 发生异常: [{}], " + MessageConstants.TASK_TRACE_ID + ": [{}]",
                    message.getPayload().getUrl(), message.getHeaders().get(MessageConstants.TASK_TRACE_ID));
            return null;
        }

        ProcessMode processMode = taskManager
                .getTaskInfoByTaskId(
                        (Integer) message.getHeaders().get(MessageConstants.TASK_ID))
                .getTask()
                .getProcessMode();
        this.log(message, res, processMode);


        if (res.getResourceStatus().equals(ResourceStatus.UNCHANGED) && ProcessMode.NORMAL.equals(processMode)) {
            return null;
        }


        return MessageBuilder
                .withPayload(res)
                .setHeader(MessageConstants.CURRENT_STAGE, StageConstants.SOURCE)
                .setHeader(MessageConstants.TASK_TRACE_ID, message.getHeaders().get(MessageConstants.TASK_TRACE_ID))
                .setHeader(MessageConstants.SOURCE_TRACE_ID, ID.generate())
                .setHeader(MessageConstants.TASK_ID, message.getHeaders().get(MessageConstants.TASK_ID))
                .setHeader(MessageConstants.TASK_NAME, message.getHeaders().get(MessageConstants.TASK_NAME))
                .setHeader(MessageConstants.SEED_URL, message.getPayload().getUrl())
                .build();


    }

    private void log(Message<HttpRequestParams> message, HttpRes res, ProcessMode processMode) {
        HttpRequestParams req = message.getPayload();

        String emoji = processMode.equals(ProcessMode.NORMAL) ? "" : "🤖";
        if (res.getStatus().is2xxSuccessful()) {
            log.info("@@ [{}{} {}] [{} {}] {} ", res.getResourceStatus().getEmoji(),
                    emoji,
                    res.getResourceStatus().getName(), req.getMethod(), res.getStatus(), req.getUrl());
        } else {
            log.warn("@@ method: [{}], status: [{}], url: [{}], request: [{}], response: [{}]",
                    req.getMethod(), res.getStatus(), req.getUrl(), req.getBody(), res.getBody());
        }
    }

    public HttpRes doHandle(Message<HttpRequestParams> message) {


        HttpRequestParams requestParams = message.getPayload();

        HttpRes httpRes = requestClient.execute(requestParams);

        String md5Digest = DigestUtils.md5DigestAsHex(httpRes.getBody().getBytes());

        ResourceStatus resourceStatus = urlMD5Service.getResourceStatus(requestParams.getUrl(), md5Digest);
        urlMD5Service.put(requestParams.getUrl(), md5Digest);

        httpRes.setResourceStatus(resourceStatus);
        return httpRes;
    }


}



